from typing import Any

from bussdcc import Event, Message, ContextProtocol
from bussdcc.io import EventSinkProtocol
from bussdcc_framework import json as framework_json


class ConsoleSink(EventSinkProtocol):
    def start(self, ctx: ContextProtocol) -> None:
        pass

    def stop(self) -> None:
        pass

    def json_fallback(self, obj: Any) -> Any:
        return framework_json.UNHANDLED

    def handle(self, evt: Event[Message]) -> None:
        if not evt.time:
            return

        record = {
            "time": evt.time,
            "type": type(evt.payload),
            "data": self.transform(evt),
        }

        try:
            line = framework_json.dumps(record, fallback=self.json_fallback)
        except Exception:
            line = repr(evt)

        print(line)

    def transform(self, evt: Event[Message]) -> Any:
        return evt.payload



from typing import Iterable

from flask_socketio import SocketIO

from bussdcc import ContextProtocol, Event, Message

from .base import FlaskApp


class BaseWebPlugin:
    name = "base"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        pass

    def event_types(self) -> Iterable[type[Message]]:
        return ()

    def handle_event(
        self,
        app: FlaskApp,
        socketio: SocketIO,
        ctx: ContextProtocol,
        evt: Event[Message],
    ) -> None:
        pass



from typing import Any, Optional, TypedDict, Iterable

from flask_socketio import SocketIO

from bussdcc import ContextProtocol
from bussdcc_framework import json as framework_json

from .base import FlaskApp
from .plugins import init_plugins
from .protocol import WebPlugin


class FlaskAppKwargs(TypedDict, total=False):
    template_folder: str
    static_folder: str


def create_app(
    ctx: ContextProtocol,
    import_name: str,
    template_folder: Optional[str] = None,
    static_folder: Optional[str] = None,
    plugins: Iterable[WebPlugin] | None = None,
) -> FlaskApp:
    kwargs: FlaskAppKwargs = {}

    if template_folder is not None:
        kwargs["template_folder"] = template_folder

    if static_folder is not None:
        kwargs["static_folder"] = static_folder

    app = FlaskApp(import_name, **kwargs)
    app.ctx = ctx
    app.socketio = SocketIO(
        app,
        cors_allowed_origins="*",
        async_mode="threading",
        json=framework_json,
    )

    @app.context_processor
    def get_context() -> dict[str, Any]:
        runtime_version = ctx.state.get("runtime.version")
        framework_version = ctx.state.get("framework.version")
        app_version = ctx.state.get("app.version")
        system_identity = ctx.state.get("system.identity")

        return dict(
            runtime_version=runtime_version,
            framework_version=framework_version,
            app_version=app_version,
            system_identity=system_identity,
        )

    init_plugins(app, ctx, plugins=plugins)

    return app



from typing import Protocol, runtime_checkable, Iterable

from flask_socketio import SocketIO

from bussdcc import ContextProtocol, Event, Message

from .base import FlaskApp


@runtime_checkable
class WebPlugin(Protocol):
    name: str

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None: ...

    def event_types(self) -> Iterable[type[Message]]: ...

    def handle_event(
        self,
        app: FlaskApp,
        socketio: SocketIO,
        ctx: ContextProtocol,
        evt: Event[Message],
    ) -> None: ...



from .base import FlaskApp
from .context import current_ctx, emit
from .interface import WebInterface
from .plugin import BaseWebPlugin
from .plugins import PluginSpec
from .protocol import WebPlugin

__all__ = [
    "FlaskApp",
    "WebInterface",
    "WebPlugin",
    "BaseWebPlugin",
    "PluginSpec",
    "current_ctx",
    "emit",
]



from typing import Any, Literal, Union, get_args, get_origin, get_type_hints
from dataclasses import fields, is_dataclass, MISSING
from datetime import date, time, datetime
from enum import Enum
from pathlib import Path
from uuid import UUID
import types

from .base import UNHANDLED


def load_atomic(tp: Any, value: Any) -> Any:
    origin = get_origin(tp)
    args = get_args(tp)

    if origin is Literal:
        if value not in args:
            raise ValueError(f"{value!r} not in {args}")
        return value

    if isinstance(tp, type) and issubclass(tp, Enum):
        if isinstance(value, tp):
            return value
        return tp(value)

    if tp is bool:
        if isinstance(value, bool):
            return value
        if value in ("true", "1", "on", "yes"):
            return True
        if value in ("false", "0", "off", "no"):
            return False
        raise TypeError(f"{tp} requires boolean-like input")

    if tp is str:
        if not isinstance(value, str):
            raise TypeError(f"{tp} requires string input")
        return value

    if tp is int:
        if isinstance(value, bool):
            raise TypeError(f"{tp} requires integer input")
        if isinstance(value, int):
            return value
        if isinstance(value, str):
            return int(value)
        raise TypeError(f"{tp} requires integer input")

    if tp is float:
        if isinstance(value, bool) or not isinstance(value, (int, float)):
            raise TypeError(f"{tp} requires numeric input")
        return float(value)

    if tp is Path:
        if not isinstance(value, (str, Path)):
            raise TypeError(f"{tp} requires path-like input")
        return Path(value)

    if tp is UUID:
        if isinstance(value, UUID):
            return value
        if isinstance(value, str):
            return UUID(value)
        raise TypeError(f"{tp} requires UUID string input")

    if tp is datetime:
        if isinstance(value, datetime):
            return value
        if isinstance(value, str):
            return datetime.fromisoformat(value)
        raise TypeError(f"{tp} requires ISO datetime string input")

    if tp is date:
        if isinstance(value, date) and not isinstance(value, datetime):
            return value
        if isinstance(value, str):
            return date.fromisoformat(value)
        raise TypeError(f"{tp} requires ISO date string input")

    if tp is time:
        if isinstance(value, time):
            return value
        if isinstance(value, str):
            return time.fromisoformat(value)
        raise TypeError(f"{tp} requires ISO time string input")

    return UNHANDLED


def load_value(tp: Any, value: Any) -> Any:
    origin = get_origin(tp)
    args = get_args(tp)

    if value is None:
        if origin in (Union, types.UnionType) and type(None) in args:
            return None
        raise TypeError(f"None is not valid for {tp!r}")

    if origin in (Union, types.UnionType):
        if type(None) in args:
            real_type = next(a for a in args if a is not type(None))
            return load_value(real_type, value)

        raise TypeError(f"{tp} is not a supported union type")

    atomic = load_atomic(tp, value)
    if atomic is not UNHANDLED:
        return atomic

    if isinstance(tp, type) and is_dataclass(tp):
        if not isinstance(value, dict):
            raise TypeError(f"{tp} requires dict input")

        type_hints = get_type_hints(tp)
        kwargs: dict[str, Any] = {}

        for f in fields(tp):
            field_tp = type_hints.get(f.name, f.type)

            if f.name in value:
                kwargs[f.name] = load_value(field_tp, value[f.name])
                continue

            if f.default is not MISSING:
                kwargs[f.name] = f.default
                continue

            if f.default_factory is not MISSING:
                kwargs[f.name] = f.default_factory()
                continue

            raise TypeError(
                f"Missing required field {f.name!r} for {tp.__module__}:{tp.__qualname__}"
            )

        return tp(**kwargs)

    if origin is list:
        if not isinstance(value, list):
            raise TypeError(f"{tp} requires list input")
        item_tp = args[0]
        return [load_value(item_tp, item) for item in value]

    if origin is tuple:
        if not isinstance(value, (list, tuple)):
            raise TypeError(f"{tp} requires list/tuple input")

        if len(args) == 2 and args[1] is Ellipsis:
            item_tp = args[0]
            return tuple(load_value(item_tp, item) for item in value)

        if len(value) != len(args):
            raise TypeError(f"Expected tuple of length {len(args)}, got {len(value)}")

        return tuple(load_value(item_tp, item) for item_tp, item in zip(args, value))

    if origin is dict:
        if not isinstance(value, dict):
            raise TypeError(f"{tp} requires dict input")

        key_tp, value_tp = args
        return {
            load_value(key_tp, k): load_value(value_tp, v) for k, v in value.items()
        }

    if origin is set:
        if not isinstance(value, list):
            raise TypeError(f"{tp} requires list input")
        item_tp = args[0]
        return {load_value(item_tp, item) for item in value}

    raise TypeError(f"{tp} is not a supported type")



import traceback
from typing import Optional

from bussdcc import Runtime as KernelRuntime
from bussdcc import Event, Message, Severity, ContextProtocol
from bussdcc.clock import ClockProtocol
from bussdcc.event import EventBusProtocol
from bussdcc.io import EventSinkProtocol
from bussdcc.state import StateStoreProtocol

from .. import message, __version__ as version


def _message_type_segments(msg: Message) -> list[str]:
    cls = type(msg)
    module_parts = cls.__module__.split(".")
    name_parts = cls.__qualname__.split(".")
    return [*module_parts, *name_parts]


class FrameworkRuntimeBase(KernelRuntime):
    def __init__(
        self,
        *,
        clock: Optional[ClockProtocol] = None,
        events: Optional[EventBusProtocol] = None,
        state: Optional[StateStoreProtocol] = None,
    ):
        super().__init__(clock=clock, events=events, state=state)
        self._sinks: list[EventSinkProtocol] = []

    def add_sink(self, sink: EventSinkProtocol) -> None:
        if self.booted:
            raise RuntimeError("Cannot add sinks after boot")
        self._sinks.append(sink)

    def boot(self) -> None:
        for sink in self._sinks:
            sink.start(self.ctx)

        super().boot()
        self.ctx.emit(message.FrameworkBooted(version=version))

    def shutdown(self, reason: Optional[str] = None) -> None:
        self.ctx.emit(message.FrameworkShuttingDown(version=version, reason=reason))

        try:
            super().shutdown(reason)
        finally:
            for sink in self._sinks:
                try:
                    sink.stop()
                except Exception:
                    pass

    def _dispatch(self, evt: Event[Message]) -> None:
        self._dispatch_to_sinks(evt)
        super()._dispatch(evt)
        self._record_runtime_telemetry(evt)

    def _dispatch_to_sinks(self, evt: Event[Message]) -> None:
        for sink in self._sinks:
            try:
                sink.handle(evt)
            except Exception as e:
                if evt.payload.severity >= Severity.ERROR:
                    continue

                self.ctx.emit(
                    message.SinkHandlerError(
                        sink=type(sink).__name__,
                        error=repr(e),
                        evt=evt,
                        traceback=traceback.format_exc(),
                    )
                )

    def _record_runtime_telemetry(self, evt: Event[Message]) -> None:
        self.ctx.state.update(
            "runtime_info.stats.message_count",
            lambda v: (v or 0) + 1,
        )

        if evt.payload.severity >= Severity.ERROR:
            self.ctx.state.update(
                "runtime_info.stats.error_count",
                lambda v: (v or 0) + 1,
            )

        self.ctx.state.update(
            f"runtime_info.severity.{evt.payload.severity.name.lower()}",
            lambda v: (v or 0) + 1,
        )

        path = "runtime_info.message_types." + ".".join(
            _message_type_segments(evt.payload)
        )
        self.ctx.state.update(path, lambda v: (v or 0) + 1)



from bussdcc import ContextProtocol

from .. import BaseWebPlugin, FlaskApp, WebPlugin

from .database import db


class SqlAlchemyPlugin(BaseWebPlugin):
    name = "sqlalchemy"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        db.init_app(app)


plugin = SqlAlchemyPlugin()



from __future__ import annotations
from dataclasses import Field, dataclass
from datetime import date, datetime, time
from enum import Enum
from types import UnionType
from typing import Any, Literal, Union, get_args, get_origin

from bussdcc_framework.metadata import FieldMetadata


def _unwrap_optional(tp: object) -> object:
    origin = get_origin(tp)
    args = get_args(tp)

    if origin in (Union, UnionType):
        non_none = [arg for arg in args if arg is not type(None)]
        if len(non_none) == 1:
            return non_none[0]

    return tp


def _enum_type(tp: object) -> type[Enum] | None:
    if isinstance(tp, type) and issubclass(tp, Enum):
        return tp
    return None


@dataclass(slots=True, frozen=True)
class FieldOption:
    value: str
    label: str


def _field_shape(
    tp: object,
    value: Any | None,
) -> tuple[str, list[FieldOption] | None, Any | None]:
    base_tp = _unwrap_optional(tp)
    origin = get_origin(base_tp)
    args = get_args(base_tp)

    enum_tp = _enum_type(base_tp)

    if origin is Literal:
        return (
            "select",
            [FieldOption(value=str(v), label=str(v)) for v in args],
            None if value is None else str(value),
        )

    if enum_tp is not None:
        if isinstance(value, Enum):
            value = value.value
        return (
            "select",
            [
                FieldOption(value=str(member.value), label=str(member.value))
                for member in enum_tp
            ],
            None if value is None else str(value),
        )

    if base_tp in (int, float):
        return "number", None, value

    if base_tp is bool:
        return "checkbox", None, value

    if base_tp is date:
        return "date", None, value

    if base_tp is time:
        return "time", None, value

    if base_tp is datetime:
        return "datetime-local", None, value

    return "text", None, value


@dataclass(slots=True, frozen=True)
class TreeField:
    name: str
    type: object
    meta: FieldMetadata
    value: Any | None = None
    input_type: str | None = None
    options: list[FieldOption] | None = None
    error: str | None = None

    @staticmethod
    def create(
        name: str,
        type: object,
        value: Any | None = None,
        *,
        label: str | None = None,
        required: bool = False,
        help: str | None = None,
        min: int | float | None = None,
        max: int | float | None = None,
        step: int | float | None = None,
        error: str | None = None,
    ) -> "TreeField":
        meta = FieldMetadata(
            label=label or name,
            required=required,
            help=help,
            min=min,
            max=max,
            step=step,
        )

        input_type, options, value = _field_shape(type, value)

        return TreeField(
            name=name,
            type=type,
            meta=meta,
            value=value,
            input_type=input_type,
            options=options,
            error=error,
        )

    @staticmethod
    def from_field(
        f: Field[object],
        value: Any | None = None,
        *,
        error: str | None = None,
        meta: FieldMetadata | None = None,
        ref_options: list[FieldOption] | None = None,
    ) -> "TreeField":
        input_type, options, value = _field_shape(f.type, value)

        if ref_options is not None:
            input_type = "select"
            options = ref_options
            if value is not None:
                value = str(value)

        return TreeField(
            name=f.name,
            type=f.type,
            meta=meta if meta is not None else FieldMetadata.from_field(f),
            value=value,
            input_type=input_type,
            options=options,
            error=error,
        )


@dataclass(slots=True, frozen=True)
class TreeMappingEntry:
    name: str
    key: TreeField
    value: TreeField | TreeNode


@dataclass(slots=True, frozen=True)
class TreeMapping:
    name: str
    meta: FieldMetadata
    entries: list[TreeMappingEntry]
    key_schema: TreeField
    value_schema: TreeField | TreeNode

    @property
    def prototype(self) -> TreeMappingEntry:
        return TreeMappingEntry(
            name="__name__",
            key=self.key_schema,
            value=self.value_schema,
        )


@dataclass(slots=True, frozen=True)
class TreeListEntry:
    name: str
    value: TreeField | TreeNode


@dataclass(slots=True, frozen=True)
class TreeList:
    name: str
    meta: FieldMetadata
    entries: list[TreeListEntry]
    item_schema: TreeField | TreeNode

    @property
    def prototype(self) -> TreeListEntry:
        return TreeListEntry(
            name="__name__",
            value=self.item_schema,
        )


@dataclass(slots=True, frozen=True)
class TreeNode:
    name: str | None
    meta: FieldMetadata
    fields: list[TreeField]
    children: list["TreeNode"]
    mappings: list[TreeMapping]
    lists: list[TreeList]
    errors: int = 0



from typing import Any, get_args, get_origin, Literal, Union
from datetime import date, time, datetime
from enum import Enum
from types import UnionType


def unwrap_optional(tp: Any) -> Any:
    origin = get_origin(tp)
    args = get_args(tp)

    if origin in (Union, UnionType):
        non_none = [arg for arg in args if arg is not type(None)]
        if len(non_none) == 1:
            return non_none[0]

    return tp


def _coerce_literal_value(options: tuple[Any, ...], raw: Any) -> Any:
    for option in options:
        if raw == option:
            return option

        if isinstance(option, bool):
            if raw in ("true", "1", "on", "yes", True) and option is True:
                return True
            if raw in ("false", "0", "off", "no", False) and option is False:
                return False
            continue

        if str(option) == str(raw):
            return option

    raise ValueError(f"{raw!r} is not one of {options!r}")


def coerce_form_value(tp: Any, raw: Any) -> Any:
    tp = unwrap_optional(tp)

    if tp is bool:
        return raw in ("on", "true", "1", "yes", True)

    if raw == "":
        if tp is str:
            return ""
        return None

    if raw is None:
        return None

    origin = get_origin(tp)

    if origin is Literal:
        return _coerce_literal_value(get_args(tp), raw)

    if isinstance(tp, type) and issubclass(tp, Enum):
        for member in tp:
            if str(member.value) == str(raw):
                return member
        raise ValueError(f"{raw!r} is not a valid value for {tp.__name__}")

    if tp is int:
        return int(raw)

    if tp is float:
        return float(raw)

    if tp in (date, time, datetime):
        return raw

    return raw



from flask import current_app

from bussdcc import ContextProtocol, Message

from .base import FlaskApp


def current_ctx() -> ContextProtocol:
    if not isinstance(current_app, FlaskApp):
        raise RuntimeError("Not inside FlaskApp context")
    return current_app.ctx


def emit(message: Message) -> None:
    current_ctx().emit(message)



from flask import Blueprint

from bussdcc import ContextProtocol

from .. import BaseWebPlugin, FlaskApp, WebPlugin

from .types import (
    FieldOption,
    TreeNode,
    TreeField,
    TreeList,
    TreeListEntry,
    TreeMapping,
    TreeMappingEntry,
)


class FormtreePlugin(BaseWebPlugin):
    name = "formtree"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        app.jinja_env.tests["tree_node"] = lambda x: isinstance(x, TreeNode)
        app.jinja_env.tests["tree_field"] = lambda x: isinstance(x, TreeField)
        app.jinja_env.tests["tree_list"] = lambda x: isinstance(x, TreeList)
        app.jinja_env.tests["tree_list_entry"] = lambda x: isinstance(x, TreeListEntry)
        app.jinja_env.tests["tree_mapping"] = lambda x: isinstance(x, TreeMapping)
        app.jinja_env.tests["tree_mapping_entry"] = lambda x: isinstance(
            x, TreeMappingEntry
        )
        app.jinja_env.tests["field_option"] = lambda x: isinstance(x, FieldOption)

        bp = Blueprint(
            "bussdcc_framework_formtree",
            __name__,
            template_folder="templates",
            static_folder="static",
            static_url_path="/_framework/formtree/static",
        )
        app.register_blueprint(bp)


plugin: WebPlugin = FormtreePlugin()



from typing import Optional
from dataclasses import dataclass

from bussdcc import Event, Message, Severity


@dataclass(slots=True, frozen=True)
class SinkHandlerError(Message):
    severity = Severity.ERROR

    sink: str
    error: str
    evt: Event[Message] | None = None
    traceback: str | None = None



from typing import Optional
from dataclasses import dataclass

from bussdcc import Message


@dataclass(slots=True, frozen=True)
class FrameworkBooted(Message):
    version: str


@dataclass(slots=True, frozen=True)
class FrameworkShuttingDown(Message):
    version: str
    reason: Optional[str]



from dataclasses import is_dataclass, fields, MISSING
from typing import Any, Optional, TypeAlias, Literal, get_origin, get_args
from enum import Enum

from bussdcc_framework.metadata import FieldMetadata

from .protocol import RefResolver
from .types import (
    TreeNode,
    TreeField,
    TreeMapping,
    TreeMappingEntry,
    TreeListEntry,
    TreeList,
)

TreeValue: TypeAlias = TreeField | TreeNode


def _is_supported_mapping_key_type(tp: object) -> bool:
    origin = get_origin(tp)

    if tp in (str, int, float, bool):
        return True

    if origin is Literal:
        return True

    if isinstance(tp, type) and issubclass(tp, Enum):
        return True

    return False


def _detect_container(tp: object) -> tuple[str | None, Any | None, Any | None]:
    origin = get_origin(tp)
    args = get_args(tp)

    if origin in (list, tuple) and args:
        item_tp = args[0]
        if len(args) == 2 and args[1] is Ellipsis:
            item_tp = args[0]
        return "list", None, item_tp

    if origin is dict and len(args) == 2:
        return "dict", args[0], args[1]

    return None, None, None


def build(
    obj: Any,
    name: Optional[str] = None,
    meta: Optional[FieldMetadata] = None,
    ref_resolver: RefResolver | None = None,
) -> TreeNode:
    if not is_dataclass(obj):
        raise TypeError("Expected dataclass")

    is_instance = not isinstance(obj, type)

    node_fields: list[TreeField] = []
    node_children: list[TreeNode] = []
    node_mappings: list[TreeMapping] = []
    node_lists: list[TreeList] = []

    if meta is None:
        meta = FieldMetadata(label=name or "Configuration")

    for f in fields(obj):
        if is_instance:
            value = getattr(obj, f.name)
        else:
            if f.default is not MISSING:
                value = f.default
            elif f.default_factory is not MISSING:
                value = f.default_factory()
            else:
                value = None

        container, key_type, subtype = _detect_container(f.type)
        field_meta = FieldMetadata.from_field(f)

        if container == "list":
            list_entries: list[TreeListEntry] = []

            list_item_schema: TreeValue
            if is_dataclass(subtype):
                list_item_schema = build(
                    subtype,
                    "value",
                    meta=FieldMetadata(label=field_meta.label),
                    ref_resolver=ref_resolver,
                )
            else:
                list_item_schema = TreeField.create("value", subtype, label="Value")

            if value:
                for i, item in enumerate(value):
                    list_entry_value: TreeValue
                    if is_dataclass(subtype):
                        list_entry_value = build(item, "value")
                    else:
                        list_entry_value = TreeField.create(
                            "value",
                            subtype,
                            value=item,
                            label="Value",
                        )

                    list_entries.append(
                        TreeListEntry(
                            name=str(i),
                            value=list_entry_value,
                        )
                    )

            node_lists.append(
                TreeList(
                    name=f.name,
                    meta=field_meta,
                    entries=list_entries,
                    item_schema=list_item_schema,
                )
            )
            continue

        if container == "dict":
            if not _is_supported_mapping_key_type(key_type):
                raise TypeError(
                    f"Unsupported mapping key type for formtree: {key_type!r}. "
                    "Only str, int, float, bool, Literal, and Enum are supported."
                )

            mapping_entries: list[TreeMappingEntry] = []

            mapping_value_schema: TreeValue
            if is_dataclass(subtype):
                mapping_value_schema = build(
                    subtype,
                    "value",
                    meta=FieldMetadata(label=field_meta.label),
                    ref_resolver=ref_resolver,
                )
            else:
                mapping_value_schema = TreeField.create(
                    "value",
                    subtype,
                    label="Value",
                )

            key_schema = TreeField.create("key", key_type, label="Key")

            if value:
                for i, (k, v) in enumerate(value.items()):
                    mapping_entry_value: TreeValue
                    if is_dataclass(subtype):
                        mapping_entry_value = build(v, "value")
                    else:
                        mapping_entry_value = TreeField.create(
                            "value",
                            subtype,
                            value=v,
                            label="Value",
                        )

                    mapping_entries.append(
                        TreeMappingEntry(
                            name=str(i),
                            key=TreeField.create("key", key_type, value=k, label="Key"),
                            value=mapping_entry_value,
                        )
                    )

            node_mappings.append(
                TreeMapping(
                    name=f.name,
                    meta=field_meta,
                    entries=mapping_entries,
                    key_schema=key_schema,
                    value_schema=mapping_value_schema,
                )
            )
            continue

        if is_dataclass(f.type):
            nested = value if is_instance else f.type
            node_children.append(
                build(
                    nested,
                    f.name,
                    meta=FieldMetadata.from_field(f),
                    ref_resolver=ref_resolver,
                )
            )
            continue

        ref_options = None

        if ref_resolver is not None and field_meta.ref is not None:
            ref_options = ref_resolver.resolve(field_meta.ref, f.type)

        node_fields.append(
            TreeField.from_field(
                f,
                value=value,
                meta=field_meta,
                ref_options=ref_options,
            )
        )

    return TreeNode(
        name=name,
        meta=meta,
        fields=node_fields,
        children=node_children,
        mappings=node_mappings,
        lists=node_lists,
    )



from flask import Blueprint

from bussdcc import ContextProtocol

from .. import BaseWebPlugin, FlaskApp, WebPlugin


class SocketIOPlugin(BaseWebPlugin):
    name = "socketio"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_framework_socketio",
            __name__,
            template_folder="templates",
            static_folder="static",
            static_url_path="/_framework/socketio/static",
        )

        app.config.setdefault("SOCKETIO_RECONNECT", True)
        app.config.setdefault("SOCKETIO_RECONNECT_ATTEMPTS", None)
        app.config.setdefault("SOCKETIO_RECONNECT_DELAY", 1000)
        app.config.setdefault("SOCKETIO_RECONNECT_DELAY_MAX", 5000)

        @app.context_processor
        def socketio_template_context() -> dict[str, object]:
            return {
                "framework_socketio": {
                    "reconnect": app.config["SOCKETIO_RECONNECT"],
                    "reconnect_attempts": app.config["SOCKETIO_RECONNECT_ATTEMPTS"],
                    "reconnect_delay": app.config["SOCKETIO_RECONNECT_DELAY"],
                    "reconnect_delay_max": app.config["SOCKETIO_RECONNECT_DELAY_MAX"],
                }
            }

        app.register_blueprint(bp)


plugin: WebPlugin = SocketIOPlugin()



from dataclasses import Field, dataclass
from types import MappingProxyType
from typing import Any


@dataclass(slots=True, frozen=True)
class FieldRef:
    kind: str | None = None
    type_: str | None = None
    protocol: str | None = None


@dataclass(slots=True, frozen=True)
class FieldMetadata:
    label: str
    group: str = "General"
    required: bool = False
    help: str | None = None
    min: int | float | None = None
    max: int | float | None = None
    step: int | float | None = None
    ref: FieldRef | None = None

    @staticmethod
    def from_field(f: Field[object]) -> "FieldMetadata":
        meta: MappingProxyType[Any, Any] = f.metadata
        raw_ref = meta.get("ref")

        ref = None
        if isinstance(raw_ref, dict):
            ref = FieldRef(
                kind=raw_ref.get("kind"),
                type_=raw_ref.get("type"),
                protocol=raw_ref.get("protocol"),
            )

        return FieldMetadata(
            label=meta.get("label", f.name),
            group=meta.get("group", "General"),
            required=meta.get("required", False),
            help=meta.get("help"),
            min=meta.get("min"),
            max=meta.get("max"),
            step=meta.get("step"),
            ref=ref,
        )



from dataclasses import replace
from typing import Any

from bussdcc_framework.codec import UNHANDLED, load_atomic

from .form import coerce_form_value, unwrap_optional
from .types import (
    TreeField,
    TreeList,
    TreeListEntry,
    TreeMapping,
    TreeMappingEntry,
    TreeNode,
)


def _child_prefix(prefix: str, child_name: str | None) -> str:
    if not child_name:
        raise RuntimeError("Collection child must have a name")
    return f"{prefix}.{child_name}" if prefix else child_name


def _collect_row_ids(flat: dict[str, Any], prefix: str) -> list[str]:
    row_ids: list[str] = []
    seen: set[str] = set()
    needle = prefix + "."

    for key in flat.keys():
        if not key.startswith(needle):
            continue

        rest = key[len(needle) :]
        if "." not in rest:
            continue

        row_id = rest.split(".", 1)[0]
        if row_id not in seen:
            seen.add(row_id)
            row_ids.append(row_id)

    return row_ids


def _is_bool_type(tp: object) -> bool:
    return unwrap_optional(tp) is bool


def _validate_field_value(field: TreeField, raw_value: Any) -> TreeField:
    if field.meta.required and raw_value in (None, ""):
        return replace(field, value=raw_value, error="This field is required")

    try:
        coerced = coerce_form_value(field.type, raw_value)
        result = load_atomic(field.type, coerced)

        if result is UNHANDLED:
            return replace(field, value=raw_value, error=None)

        return replace(field, value=raw_value, error=None)
    except Exception as e:
        return replace(field, value=raw_value, error=str(e))


def _validate_node(node: TreeNode, flat: dict[str, Any], prefix: str = "") -> TreeNode:
    node_prefix = prefix
    if node.name:
        node_prefix = f"{prefix}.{node.name}" if prefix else node.name

    fields: list[TreeField] = []
    children: list[TreeNode] = []
    lists: list[TreeList] = []
    mappings: list[TreeMapping] = []
    errors = 0

    for field in node.fields:
        key = f"{node_prefix}.{field.name}" if node_prefix else field.name

        if _is_bool_type(field.type):
            raw_value = flat.get(key)
        else:
            raw_value = flat.get(key, field.value)

        validated = _validate_field_value(field, raw_value)
        fields.append(validated)

        if validated.error:
            errors += 1

    for child in node.children:
        validated_child = _validate_node(child, flat, node_prefix)
        children.append(validated_child)
        errors += validated_child.errors

    for tree_list in node.lists:
        prefix_for_list = _child_prefix(node_prefix, tree_list.name)
        row_ids = _collect_row_ids(flat, prefix_for_list)

        list_entries: list[TreeListEntry] = []

        for row_id in row_ids:
            row_prefix = f"{prefix_for_list}.{row_id}"
            validated_item: TreeField | TreeNode

            if isinstance(tree_list.item_schema, TreeField):
                value_key = f"{row_prefix}.value"
                raw_value = (
                    flat.get(value_key)
                    if _is_bool_type(tree_list.item_schema.type)
                    else flat.get(value_key, tree_list.item_schema.value)
                )
                validated_item = _validate_field_value(
                    tree_list.item_schema,
                    raw_value,
                )
                if validated_item.error:
                    errors += 1

            elif isinstance(tree_list.item_schema, TreeNode):
                validated_item = _validate_node(tree_list.item_schema, flat, row_prefix)
                errors += validated_item.errors

            else:
                raise RuntimeError("List missing item schema")

            list_entries.append(TreeListEntry(name=row_id, value=validated_item))

        lists.append(
            TreeList(
                name=tree_list.name,
                meta=tree_list.meta,
                entries=list_entries,
                item_schema=tree_list.item_schema,
            )
        )

    for mapping in node.mappings:
        prefix_for_mapping = _child_prefix(node_prefix, mapping.name)
        row_ids = _collect_row_ids(flat, prefix_for_mapping)

        mapping_entries: list[TreeMappingEntry] = []

        for row_id in row_ids:
            row_prefix = f"{prefix_for_mapping}.{row_id}"

            raw_key = flat.get(f"{row_prefix}.key")
            validated_key = _validate_field_value(mapping.key_schema, raw_key)

            validated_value: TreeField | TreeNode

            if isinstance(mapping.value_schema, TreeField):
                value_key = f"{row_prefix}.value"
                raw_value = (
                    flat.get(value_key)
                    if _is_bool_type(mapping.value_schema.type)
                    else flat.get(value_key, mapping.value_schema.value)
                )
                validated_value = _validate_field_value(
                    mapping.value_schema,
                    raw_value,
                )
                if validated_value.error:
                    errors += 1

            elif isinstance(mapping.value_schema, TreeNode):
                validated_value = _validate_node(mapping.value_schema, flat, row_prefix)
                errors += validated_value.errors

            else:
                raise RuntimeError("Mapping missing value schema")

            if validated_key.error:
                errors += 1

            mapping_entries.append(
                TreeMappingEntry(
                    name=row_id,
                    key=validated_key,
                    value=validated_value,
                )
            )

        mappings.append(
            TreeMapping(
                name=mapping.name,
                meta=mapping.meta,
                entries=mapping_entries,
                key_schema=mapping.key_schema,
                value_schema=mapping.value_schema,
            )
        )

    return TreeNode(
        name=node.name,
        meta=node.meta,
        fields=fields,
        children=children,
        mappings=mappings,
        lists=lists,
        errors=errors,
    )


def validate(tree: TreeNode, flat: dict[str, Any]) -> TreeNode:
    return _validate_node(tree, flat)



from typing import Iterable, Any
from importlib.metadata import entry_points

from bussdcc import ContextProtocol

from .base import FlaskApp
from .protocol import WebPlugin

ENTRYPOINT_GROUP = "bussdcc_framework.web"
PluginSpec = str | WebPlugin


def _load_entrypoint_plugin(name: str) -> WebPlugin:
    matches = entry_points().select(group=ENTRYPOINT_GROUP, name=name)
    ep = next(iter(matches), None)

    if ep is None:
        raise RuntimeError(f"Unknown web plugin: {name}")

    try:
        obj: Any = ep.load()
    except Exception as e:
        raise RuntimeError(f"Failed to load web plugin {name!r}: {e}") from e

    if not isinstance(obj, WebPlugin):
        raise RuntimeError(
            f"Web plugin ignored: {name}: object does not implement WebPlugin"
        )

    return obj


def _resolve_plugin(spec: PluginSpec) -> WebPlugin:
    if isinstance(spec, str):
        return _load_entrypoint_plugin(spec)
    return spec


def resolve_plugins(
    plugins: Iterable[PluginSpec] | None = None,
) -> list[WebPlugin]:
    seen: set[str] = set()
    resolved: list[WebPlugin] = []

    if plugins is None:
        return resolved

    for spec in plugins:
        plugin = _resolve_plugin(spec)

        if plugin.name in seen:
            continue

        resolved.append(plugin)
        seen.add(plugin.name)

    return resolved


def init_plugins(
    app: FlaskApp,
    ctx: ContextProtocol,
    plugins: Iterable[WebPlugin] | None = None,
) -> None:
    if plugins is None:
        return

    for plugin in plugins:
        plugin.init_app(app, ctx)



import json
from typing import Any

from .codec import UNHANDLED, dump_value
from .codec.dump import DumpFallback


def to_jsonable(obj: Any, fallback: DumpFallback | None = None) -> Any:
    return dump_value(obj, fallback=fallback)


def dumps(obj: Any, **kwargs: Any) -> str:
    fallback = kwargs.pop("fallback", None)
    kwargs.setdefault("separators", (",", ":"))
    return json.dumps(to_jsonable(obj, fallback=fallback), **kwargs)


def loads(s: str, **kwargs: Any) -> Any:
    return json.loads(s, **kwargs)


def to_json_primitives(obj: Any, fallback: DumpFallback | None = None) -> Any:
    return loads(dumps(obj, fallback=fallback))


__all__ = [
    "UNHANDLED",
    "to_jsonable",
    "to_json_primitives",
    "dumps",
    "loads",
]



from typing import Protocol

from bussdcc_framework.metadata import FieldRef

from .types import FieldOption


class RefResolver(Protocol):
    def resolve(
        self, ref: FieldRef, field_type: object
    ) -> list[FieldOption] | None: ...



from typing import Iterator
from datetime import datetime
from pathlib import Path
from importlib import import_module
import json

from bussdcc import Event, Message
from bussdcc.io import EventSourceProtocol

from bussdcc_framework.codec import load_value


class JsonlSource(EventSourceProtocol):
    def __init__(self, root: str | Path) -> None:
        self.root = Path(root)
        self.start_at = None

    def __iter__(self) -> Iterator[Event[Message]]:
        files = sorted(self.root.rglob("*.jsonl"))
        for file in files:
            with file.open() as f:
                for line in f:
                    try:
                        record = json.loads(line)
                    except json.JSONDecodeError as e:
                        print(f"{file}: {e}")
                        continue

                    evt_time = datetime.fromisoformat(record["time"])

                    if self.start_at and evt_time < self.start_at:
                        continue

                    if not "type" in record:
                        continue  # invalid format

                    message_type = record["type"]
                    module_path, object_name = message_type.split(":", 2)
                    module = import_module(module_path)
                    message_cls = getattr(module, object_name)

                    yield Event(
                        time=evt_time,
                        payload=load_value(message_cls, record["data"]),
                    )



from flask import Blueprint

from bussdcc import ContextProtocol

from .. import BaseWebPlugin, FlaskApp, WebPlugin


class BootstrapPlugin(BaseWebPlugin):
    name = "bootstrap"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_framework_bootstrap",
            __name__,
            template_folder="templates",
            static_folder="static",
            static_url_path="/_framework/bootstrap/static",
        )
        app.register_blueprint(bp)


plugin = BootstrapPlugin()



from typing import Any

from .types import TreeNode, TreeField
from .form import coerce_form_value, unwrap_optional


def _child_prefix(prefix: str, child_name: str | None) -> str:
    if not child_name:
        raise RuntimeError("Collection child must have a name")
    return f"{prefix}.{child_name}" if prefix else child_name


def _collect_row_ids(flat: dict[str, Any], prefix: str) -> list[str]:
    row_ids: list[str] = []
    seen: set[str] = set()
    needle = prefix + "."

    for key in flat.keys():
        if not key.startswith(needle):
            continue

        rest = key[len(needle) :]
        if "." not in rest:
            continue

        row_id = rest.split(".", 1)[0]
        if row_id not in seen:
            seen.add(row_id)
            row_ids.append(row_id)

    return row_ids


def _is_bool_type(tp: object) -> bool:
    return unwrap_optional(tp) is bool


def unflatten(tree: TreeNode, flat: dict[str, Any]) -> dict[str, Any]:
    def build(node: TreeNode, prefix: str = "") -> dict[str, Any]:
        result: dict[str, Any] = {}

        node_prefix = prefix
        if node.name:
            node_prefix = f"{prefix}.{node.name}" if prefix else node.name

        for field in node.fields:
            key = f"{node_prefix}.{field.name}" if node_prefix else field.name

            if _is_bool_type(field.type):
                value = flat.get(key)
            else:
                value = flat.get(key, field.value)

            result[field.name] = coerce_form_value(field.type, value)

        for child in node.children:
            if not child.name:
                raise RuntimeError("Nested child must have a name")
            result[child.name] = build(child, node_prefix)

        for tree_list in node.lists:
            prefix_for_list = _child_prefix(node_prefix, tree_list.name)
            row_ids = _collect_row_ids(flat, prefix_for_list)

            items: list[Any] = []
            for row_id in row_ids:
                row_prefix = f"{prefix_for_list}.{row_id}"

                if isinstance(tree_list.item_schema, TreeField):
                    value_key = f"{row_prefix}.value"
                    if _is_bool_type(tree_list.item_schema.type):
                        raw_value = flat.get(value_key)
                    else:
                        raw_value = flat.get(value_key, tree_list.item_schema.value)

                    items.append(
                        coerce_form_value(tree_list.item_schema.type, raw_value)
                    )
                elif isinstance(tree_list.item_schema, TreeNode):
                    items.append(build(tree_list.item_schema, row_prefix))
                else:
                    raise RuntimeError("List missing item schema")

            result[tree_list.name] = items

        for mapping in node.mappings:
            prefix_for_mapping = _child_prefix(node_prefix, mapping.name)
            row_ids = _collect_row_ids(flat, prefix_for_mapping)

            dict_items: dict[Any, Any] = {}
            for row_id in row_ids:
                row_prefix = f"{prefix_for_mapping}.{row_id}"

                raw_dict_key = flat.get(f"{row_prefix}.key")
                if raw_dict_key in (None, ""):
                    continue

                dict_key = coerce_form_value(mapping.key_schema.type, raw_dict_key)
                if dict_key is None:
                    continue

                if isinstance(mapping.value_schema, TreeField):
                    value_key = f"{row_prefix}.value"
                    if _is_bool_type(mapping.value_schema.type):
                        raw_value = flat.get(value_key)
                    else:
                        raw_value = flat.get(value_key, mapping.value_schema.value)

                    dict_items[dict_key] = coerce_form_value(
                        mapping.value_schema.type,
                        raw_value,
                    )
                elif isinstance(mapping.value_schema, TreeNode):
                    dict_items[dict_key] = build(mapping.value_schema, row_prefix)
                else:
                    raise RuntimeError("Mapping missing value schema")

            result[mapping.name] = dict_items

        return result

    return build(tree)



from typing import Optional, Iterable
import threading

from werkzeug.serving import make_server
from flask_socketio import SocketIO

from bussdcc import Process, ContextProtocol, Event, Message
from bussdcc_framework import message

from .base import FlaskApp
from .factory import create_app
from .plugins import PluginSpec, resolve_plugins
from .protocol import WebPlugin


class WebInterface(Process):
    name = "web"

    def __init__(
        self,
        import_name: Optional[str] = None,
        host: str = "127.0.0.1",
        port: int = 5000,
        template_folder: Optional[str] = None,
        static_folder: Optional[str] = None,
        plugins: Iterable[PluginSpec] | None = None,
    ) -> None:
        self.import_name = import_name or __name__
        self.template_folder = template_folder
        self.static_folder = static_folder
        self.plugins = list(plugins) if plugins is not None else []
        self.host = host
        self.port = port
        self._thread: threading.Thread | None = None
        self._plugins: list[WebPlugin] = []
        self._event_handlers: dict[type[Message], list[WebPlugin]] = {}

    def iter_plugins(self) -> Iterable[PluginSpec]:
        return []

    def register_routes(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        pass

    def register_socketio(self, socketio: SocketIO, ctx: ContextProtocol) -> None:
        pass

    def start(self, ctx: ContextProtocol) -> None:
        plugin_specs = [*self.iter_plugins(), *self.plugins]
        self._plugins = resolve_plugins(plugin_specs)
        self._event_handlers = self._build_event_handlers(self._plugins)

        self.app = create_app(
            ctx,
            self.import_name,
            self.template_folder,
            self.static_folder,
            plugins=self._plugins,
        )
        self.socketio = self.app.socketio

        self.register_routes(self.app, ctx)
        self.register_socketio(self.socketio, ctx)

        self._thread = threading.Thread(
            target=self._run,
            name=self.name,
            daemon=True,
        )
        self._thread.start()

        ctx.emit(message.WebInterfaceStarted(host=self.host, port=self.port))

    def _build_event_handlers(
        self,
        plugins: Iterable[WebPlugin],
    ) -> dict[type[Message], list[WebPlugin]]:
        handlers: dict[type[Message], list[WebPlugin]] = {}

        for plugin in plugins:
            for event_type in plugin.event_types():
                handlers.setdefault(event_type, []).append(plugin)

        return handlers

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        app = getattr(self, "app", None)
        socketio = getattr(self, "socketio", None)

        if app is None or socketio is None:
            return

        seen: set[str] = set()

        for cls in type(evt.payload).__mro__:
            if cls is object:
                continue

            for plugin in self._event_handlers.get(cls, ()):
                if plugin.name in seen:
                    continue

                plugin.handle_event(app, socketio, ctx, evt)
                seen.add(plugin.name)

    def _run(self) -> None:
        self._server = make_server(
            host=self.host,
            port=self.port,
            app=self.app,
            threaded=True,
        )

        self._server.serve_forever()

    def stop(self, ctx: ContextProtocol) -> None:
        server = getattr(self, "_server", None)
        if server:
            server.shutdown()

        if self._thread:
            self._thread.join(timeout=5)



from typing import Any, Callable
from dataclasses import fields, is_dataclass
from datetime import date, time, datetime
from enum import Enum
from pathlib import Path
from uuid import UUID

from .base import UNHANDLED

DumpFallback = Callable[[Any], Any]


def dump_value(obj: Any, fallback: DumpFallback | None = None) -> Any:
    if isinstance(obj, dict):
        return {k: dump_value(v, fallback=fallback) for k, v in obj.items()}

    if isinstance(obj, list):
        return [dump_value(v, fallback=fallback) for v in obj]

    if isinstance(obj, tuple):
        return [dump_value(v, fallback=fallback) for v in obj]

    if isinstance(obj, set):
        return [dump_value(v, fallback=fallback) for v in obj]

    if obj is None or isinstance(obj, (str, int, float, bool)):
        return obj

    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()

    if isinstance(obj, Enum):
        return obj.value

    if isinstance(obj, Path):
        return str(obj)

    if isinstance(obj, UUID):
        return str(obj)

    if is_dataclass(obj) and not isinstance(obj, type):
        return {
            f.name: dump_value(getattr(obj, f.name), fallback=fallback)
            for f in fields(obj)
        }

    if isinstance(obj, type):
        return f"{obj.__module__}:{obj.__qualname__}"

    if fallback is not None:
        alt = fallback(obj)
        if alt is not UNHANDLED:
            return dump_value(alt, fallback=fallback)

    raise TypeError(f"{type(obj)} not serializable")



from pathlib import Path
from datetime import datetime, timezone, timedelta
from typing import Any, TextIO
import threading

from bussdcc import Event, Message, ContextProtocol
from bussdcc.io import EventSinkProtocol
from bussdcc_framework import json as framework_json


class JsonlSink(EventSinkProtocol):
    def __init__(
        self,
        root: str | Path,
        interval: float = 600.0,
    ):
        self.root = Path(root)
        self.interval = timedelta(seconds=interval)

        self._file: TextIO | None = None
        self._current_segment_start: datetime | None = None
        self._lock = threading.Lock()

    def start(self, ctx: ContextProtocol) -> None:
        self.root.mkdir(parents=True, exist_ok=True)

    def stop(self) -> None:
        if self._file:
            self._file.close()
            self._file = None

    def json_fallback(self, obj: Any) -> Any:
        return framework_json.UNHANDLED

    def handle(self, evt: Event[Message]) -> None:
        if not evt.time:
            return

        segment_start = self._segment_start(evt.time)

        with self._lock:
            if segment_start != self._current_segment_start:
                self._rotate(segment_start)

            record = {
                "time": evt.time,
                "type": type(evt.payload),
                "data": self.transform(evt),
            }

            line = framework_json.dumps(record, fallback=self.json_fallback)
            assert self._file is not None
            self._file.write(line + "\n")

    def transform(self, evt: Event[Message]) -> Any:
        return evt.payload

    def _segment_start(self, dt: datetime) -> datetime:
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)

        interval_seconds = self.interval.total_seconds()
        timestamp = dt.timestamp()
        bucket = int(timestamp // interval_seconds) * interval_seconds
        return datetime.fromtimestamp(bucket, tz=dt.tzinfo)

    def _rotate(self, segment_start: datetime) -> None:
        if self._file:
            self._file.close()

        self._current_segment_start = segment_start

        day_dir = self.root / segment_start.strftime("%Y-%m-%d")
        day_dir.mkdir(parents=True, exist_ok=True)

        filename = segment_start.strftime("%H-%M-%S.jsonl")
        path = day_dir / filename

        self._file = path.open("a", buffering=1)



from .interface.web import (
    FlaskApp,
    WebInterface,
    WebPlugin,
    BaseWebPlugin,
    PluginSpec,
    current_ctx,
    emit,
)

__all__ = [
    "FlaskApp",
    "WebInterface",
    "WebPlugin",
    "BaseWebPlugin",
    "PluginSpec",
    "current_ctx",
    "emit",
]
